import json
import logging
import os
import random
import re
import shutil
import subprocess
from os import path
from pathlib import Path
from urllib.parse import urlparse

import httpx
import requests

from mtmlib.mtutils import bash

logger = logging.getLogger()


def gitParseOwnerRepo(giturl: str):
    """再git url 中识别出owner 和 repo 名称"""
    parsed = urlparse(giturl)
    path = parsed.path
    m = re.match(r"/(.*)/(.*).git(.*)", path, flags=0)
    if m:
        owner = m.groups()[0]
        repo = m.groups()[1]
        file = m.groups()[2]
        return (owner, repo, file)
    return None


def gitup(giturl: str):
    """
    给定git 网址，下载到指定路径并根据规则运行相关代码。
    """
    parsed = urlparse(giturl)
    owner, repo, file = gitParseOwnerRepo(giturl)

    clone_to = path.join(GITCLONE_BASEDIR, repo)
    gitclone(owner, repo, parsed.username, clone_to)
    if not file:
        logger.info("no entry script,skip launch")
    if file:
        file = file.lstrip("/")
        scriptFile = path.join(clone_to, file)
        if not Path(scriptFile).exists():
            logger.warning(f"入口文件不存在{scriptFile}")

        logpath = path.join(clone_to, GITUP_LOGFILE)
        Path(scriptFile).chmod(0o700)
        logger.info(f"""========================================
[ ✅ gitup]
entry script\t: {scriptFile}
logs\t: tail -f {logpath}
========================================
""")
        logfile = open(logpath, "w")
        subprocess.Popen(
            [scriptFile],
            stdout=logfile,  # 输出重定向到文件。
            stderr=logfile,
            # stdout=subprocess.PIPE,
            # stderr=subprocess.PIPE,
            cwd=clone_to,
            shell=False,
        )
    return logpath


def gitCopy(giturl: str, dest_dir: str):
    """
    以类似cp命令的方式，复制仓库中的文件到指定目录，
    为什么不直接用git clone? 因为如果目前文件夹非空，会运行失败
    """
    parsed = urlparse(giturl)
    owner, repo, file = gitParseOwnerRepo(giturl)
    tmp_dir = "/tmp/gitclone_" + str(random.randint(0, 999999999))
    clone_to = path.join(tmp_dir, repo)
    gitclone(owner, repo, parsed.username, clone_to)
    logger.info(f"copytree from `{clone_to}` to `{dest_dir}` ")
    shutil.copytree(clone_to, dest_dir, dirs_exist_ok=True)


def dockerdev_git(giturl: str):
    """在容器内clone 源码并启动为基于容器的开发环境"""
    parsed = urlparse(giturl)
    owner, repo, file = gitParseOwnerRepo(giturl)
    clone_to = path.join(repo)
    logger.info(f"clone 到 {clone_to}")
    gitclone(owner, repo, parsed.username, clone_to)


def gitclone(owner, repo, token, dest_dir):
    """克隆一个github仓库"""
    _token = token or os.environ.get("GITTOKEN")

    bash(f"rm -rdf {dest_dir}")
    cmd = (
        f"git clone --depth=1 https://{_token}@github.com/{owner}/{repo}.git {dest_dir}"
    )
    logger.info(f"cmd {cmd}")
    cp = bash(cmd)
    if cp.returncode != 0:
        logger.info(f"clone 失败 {owner}, {repo},{dest_dir}")
        logger.info(cp.stderr)
    else:
        logger.info("clone 成功")


def deploy_repo(repo_url):
    """更新源码，目前配合github hook 的功能，暂时写死
    TODO: 目前仅支持ssh的方式拉取源码。要更新到支持github token
    """
    # os.system('curl -q --insecure https://116.202.120.181/api/ip')
    # os.system('curl -q http://google.com')
    url = urlparse(repo_url)
    owner = url.path.split("/")[1]
    repo_name = url.path.split("/")[2].rstrip(".git")

    targetdir = "/deploy/" + owner + "/" + repo_name
    if os.path.exists(targetdir):
        logger.info("文件夹 %s 存在，拉取github" % targetdir)
        os.system("cd %s && git reset --hard origin/main && git pull" % targetdir)
    else:
        parent_dir = Path(targetdir).parent
        Path(parent_dir).mkdir(mode=0o777, parents=True, exist_ok=True)
        logger.info("部署路径 %s" % targetdir)
        os.system(f"git clone {repo_url} {targetdir}")
    # 运行
    is_entry_exists = False
    for entry_file in DEFAULT_ENTRY_SCRIPTS.split(" "):
        fullpath = os.path.join(targetdir, entry_file)
        logger.info(f"搜索入口脚本；{fullpath}")
        if os.path.exists(fullpath):
            is_entry_exists = True
            logger.info("执行脚本 %s " % targetdir)
            proc = subprocess.Popen(["sh", "-c", fullpath], cwd=targetdir)
            break
    if not is_entry_exists:
        logger.warning(f"仓库入库脚本文件'{DEFAULT_ENTRY_SCRIPTS}'不存在，跳过启动")


class GHRepo:
    """对一个github repo进行操作"""

    def __init__(
        self, url: str = None, token: str = None, owner: str = None, repo: str = None
    ):
        if url:
            repourl = urlparse(url)
            self.token = repourl.password
            self.owner = repourl.username
            pathitems = repourl.path.split("/")
            self.repo = pathitems[1]
            print("REPO")
            print(
                f"根据网址否则ghrepo， token {self.token}, owner:{self.owner}, repo:{self.repo}"
            )
        else:
            self.token = token
            self.owner = owner
            self.repo = repo

        self.http_headers = {
            "Authorization": f"token {self.token}",
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36 QIHU 360SE",
        }

    def repoInfo(self):
        """获取当前仓库的信息"""
        r = requests.get(
            f"https://api.github.com/orgs/{self.owner}/repos",
            data=payload,
            headers=self.http_headers,
        )
        # text = r.text
        print("DEBUG 当前仓库信息", json.loads(r.text))

    def _put(apiurl, payload):
        """发出put请求"""
        return requests.put(
            f"https://api.github.com/orgs/{self.owner}/repos",
            data=payload,
            headers=self.http_headers,
        )

    # 将文件转换为base64编码，上传文件必须将文件以base64格式上传

    def delete_file(self, file_path, sha, message="api delete"):
        url = f"https://api.github.com/repos/{self.owner}/{self.repo}/contents/{file_path}"
        headers = {"Authorization": "token " + self.token}
        payload = json.dumps(
            {
                "sha": sha,
                "message": message,
            }
        )
        resp = requests.delete(url=url, data=payload, headers=headers)
        json_data = resp.json()
        # print("删除结果：")
        # print(json_data)
        message = json_data.get("message")
        if message == "Not Found":
            return None
        return json_data

    def get_file_content(self, file_path: str):
        url = f"https://api.github.com/repos/{self.owner}/{self.repo}/contents/{file_path}"
        headers = {"Authorization": "token " + self.token}
        resp = requests.get(url=url, data=None, headers=headers)
        return resp.json()

    def _write_content(self, file_path: str, file_data: bytes):
        req = requests.put(
            url=f"https://api.github.com/repos/{self.owner}/{self.repo}/contents/{file_path}",
            data=json.dumps(
                {
                    "message": "message",
                    "committer": {"name": "[user]", "email": "user@163.com"},
                    "content": file_base64(file_data),
                }
            ),
            headers={"Authorization": "token " + self.token},
        )
        req.encoding = "utf-8"
        return json.loads(req.text)

    # 上传文件（单个）
    def write_file_content(self, file_path: str, file_data: bytes, skipExists=False):
        """以覆盖的方式写入文件到github仓库"""
        contentJson = self.get_file_content(file_path)
        sha = contentJson.get("sha")
        if sha and not skipExists:
            print(f"删除旧文件：{file_path}")
            self.delete_file(file_path, sha=sha)
            self._write_content(file_path, file_data)
        elif sha and skipExists:
            print(f"跳过{file_path}")
        else:
            self._write_content(file_path, file_data)

    def write_file_content_mutil(self, files=[]):
        """以覆盖的方式写入文件到github仓库"""
        for file in files:
            print(f"上传：{file}")
            self.write_file_content(file["path"], file["content"])

        # print(json.dumps(re_data))
        # print(f"https://cdn.jsdelivr.net/gh/{self.owner}/{self.repo}/{file_path}")


def git_commit_push(commit_msg: str = "auto-commit"):
    bash(f"""git add -A
	git commit -m "{commit_msg}" --allow-empty
	# git push origin main --force
    git push --follow-tags
""")


##########################################################################################################
# 新版代码开始
async def get_github_user_data(access_token: str):
    # 使用 access token 获取用户信息
    async with httpx.AsyncClient() as client:
        user_response = await client.get(
            "https://api.github.com/user",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        user_data = user_response.json()
        return user_data


async def git_clone_repos(*, base_dir: str, repo_urls: list[str], force: bool = False):
    # 将多个指定的 git 开源项目 clone 到指定的目类下
    # all_git_srcs = ["https://github.com/Skyvern-AI/skyvern"]
    for gitsrc in repo_urls:
        uri = urlparse(gitsrc)
        sub_dir = uri.path.split("/")[-1]
        target_dir = Path(base_dir).joinpath(sub_dir)
        if Path(target_dir).exists() and not force:
            logger.info("⚠ %s 已存在", target_dir)
        else:
            if Path(target_dir).exists():
                logger.info("⚠ %s 已存在，删除", target_dir)
                shutil.rmtree(target_dir)
            cmd = f"git clone --depth=1 {gitsrc} {target_dir}"
            bash(cmd)



import logging


class LoggingOptions:
    format: str | None = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    level: str | None = "info"


def setup_logging(option: LoggingOptions):
    log_format = option.format or "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    logging.basicConfig(
        level=getattr(logging, option.level.upper(), logging.INFO),
        format=log_format,
        handlers=[
            logging.StreamHandler(),
        ],
    )

    # root_logger = logging.getLogger()
    # log_file = settings.LOGGING_PATH
    # if log_file:
    #     file_handler = logging.FileHandler(log_file)
    #     file_handler.setLevel(root_logger.level)
    #     file_handler.setFormatter(logging.Formatter(log_format))
    #     root_logger.addHandler(file_handler)

    # if settings.LOKI_ENDPOINT:
    #     print(
    #         f"use loki logging handler: {settings.LOKI_USER},{settings.LOKI_ENDPOINT}"
    #     )
    #     if not settings.GRAFANA_TOKEN:
    #         print("missing GRAFANA_TOKEN, skip setup loki")
    #     else:
    #         import logging_loki

    #         handler = logging_loki.LokiHandler(
    #             url=settings.LOKI_ENDPOINT,
    #             tags={
    #                 "application": settings.app_name,
    #                 "deploy": settings.otel_deploy_name,
    #             },
    #             auth=(settings.LOKI_USER, settings.GRAFANA_TOKEN),
    #             version="1",
    #         )
    #         root_logger.addHandler(handler)

    # if settings.IS_TRACE_HTTPX:
    #     httpx_logger = logging.getLogger("httpx")
    #     httpx_logger.setLevel(logging.INFO)



import gzip
import importlib
import os
import random
import re
import shutil
import subprocess
import sys
import urllib
from io import BytesIO
from pathlib import Path

import httpx
from nanoid import generate


def gen_orm_id_key():
    return generate("1234567890abcdef", 16)


def ranstr(num):
    H = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
    ret = []
    for i in range(num):
        ret.append(random.choice(H))
    return "".join(ret)


async def write_to_file(filename: str, text: str) -> None:
    """Asynchronously write text to a file in UTF-8 encoding.

    Args:
        filename (str): The filename to write to.
        text (str): The text to write.
    """
    import aiofiles

    # Ensure text is a string
    if not isinstance(text, str):
        text = str(text)

    # Convert text to UTF-8, replacing any problematic characters
    text_utf8 = text.encode("utf-8", errors="replace").decode("utf-8")

    async with aiofiles.open(filename, "w", encoding="utf-8") as file:
        await file.write(text_utf8)


async def write_text_to_md(text: str, filename: str = "") -> str:
    """Writes text to a Markdown file and returns the file path.

    Args:
        text (str): Text to write to the Markdown file.

    Returns
    -------
        str: The file path of the generated Markdown file.
    """
    file_path = f"outputs/{filename[:60]}.md"
    await write_to_file(file_path, text)
    return urllib.parse.quote(file_path)


async def write_md_to_pdf(text: str, filename: str = "") -> str:
    """Converts Markdown text to a PDF file and returns the file path.

    Args:
        text (str): Markdown text to convert.

    Returns
    -------
        str: The encoded file path of the generated PDF.
    """
    file_path = f"outputs/{filename[:60]}.pdf"

    try:
        from md2pdf.core import md2pdf

        md2pdf(
            file_path,
            md_content=text,
            # md_file_path=f"{file_path}.md",
            css_file_path="./frontend/pdf_styles.css",
            base_url=None,
        )
        print(f"Report written to {file_path}.pdf")
    except Exception as e:
        print(f"Error in converting Markdown to PDF: {e}")
        return ""

    encoded_file_path = urllib.parse.quote(file_path)
    return encoded_file_path


async def write_md_to_word(text: str, filename: str = "") -> str:
    """Converts Markdown text to a DOCX file and returns the file path.

    Args:
        text (str): Markdown text to convert.

    Returns
    -------
        str: The encoded file path of the generated DOCX.
    """
    import mistune

    file_path = f"outputs/{filename[:60]}.docx"

    try:
        from docx import Document
        from htmldocx import HtmlToDocx

        # Convert report markdown to HTML
        html = mistune.html(text)
        # Create a document object
        doc = Document()
        # Convert the html generated from the report to document format
        HtmlToDocx().add_html_to_document(html, doc)

        # Saving the docx document to file_path
        doc.save(file_path)

        print(f"Report written to {file_path}")

        encoded_file_path = urllib.parse.quote(file_path)
        return encoded_file_path

    except Exception as e:
        print(f"Error in converting Markdown to DOCX: {e}")
        return ""


# Regular expression for a valid filename
VALID_FILENAME_RE = r"^[\w,\s-]+\.[A-Za-z]{2,}$"


def is_valid_filename(filename: str) -> bool:
    return re.match(VALID_FILENAME_RE, filename) is not None


def ensure_directory_exists(directory: str):
    p = Path(directory)
    if not p.exists():
        Path.mkdir(directory, parents=True)


async def write_file(path: str, content: str):
    ensure_directory_exists(Path(path).parent)
    with Path.open(path, "w") as r:
        r.write(content)


async def pipe(reader, writer):
    """流转发"""
    while True:
        if reader.at_eof():
            return
        data = await reader.read(4096)
        if len(data) > 0:
            print(">:", len(data))
            writer.write(data)


def is_in_gitpod():
    """是否在gitpod中运行"""
    return bool(os.environ.get("GITPOD_REPO_ROOT"))


def in_colab():
    return bool(os.environ.get("COLAB_RELEASE_TAG"))


def get_pyproject_version():
    # content = Path("./pyproject.toml").read_text()
    content = Path("./mtmai/core/__version__.py").read_text()
    match = re.search(r'version\s*=\s*"(\d+\.\d+\.\d+)"', content)
    if not match:
        msg = "版本号未在 pyproject.toml 中找到"
        raise ValueError(msg)
    return match.group(1)


def increment_patch_version(version):
    major, minor, patch = map(int, version.split("."))
    patch += 1
    return f"{major}.{minor}.{patch}"


def pyproject_patch_version(project_dir: str = "."):
    pyproject_path = Path(f"{project_dir}/pyproject.toml")
    content = pyproject_path.read_text(encoding="utf-8")

    old_version = get_pyproject_version(project_dir)
    new_version = increment_patch_version(old_version)

    # 使用正则表达式替换旧版本号为新版本号
    new_content = re.sub(
        r'version\s*=\s*"\d+\.\d+\.\d+"', f'version = "{new_version}"', content
    )
    pyproject_path.write_text(new_content, encoding="utf-8")

    content2 = f'version = "{new_version}"\n'
    Path("./mtmai/core/__version__.py").write_text(content2)


def download_and_extract_gz(url, output_file_path):
    """
    下载并直接解压 .gz 文件到指定文件路径

    :param url: 文件下载链接
    :param output_file_path: 输出文件的完整路径
    """
    dest_folder = Path(output_file_path)

    if not Path.exists(dest_folder):
        Path.mkdir(dest_folder, parents=True)

    response = httpx.get(url, stream=True)
    response.raise_for_status()
    with gzip.open(BytesIO(response.content), "rb") as f_in:  # noqa: SIM117
        with open(output_file_path, "wb") as f_out:  # noqa: PTH123
            shutil.copyfileobj(f_in, f_out)


def install_and_import(package):
    try:
        __import__(package)
    except ModuleNotFoundError:
        print(f"Module '{package}' not found. Installing...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        # Re-import the package after installation
        globals()[package] = __import__(package)


def command_exists(cmd_name: str) -> bool:
    try:
        # Run the command with 'which' to check if it exists
        result = subprocess.run(
            ["which", cmd_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        return result.returncode == 0
    except Exception as e:
        print(f"An error occurred: {e}")
        return False

    # 方式2:
    # from distutils.spawn import find_executable
    # return find_executable(cmd_name) is not None


def bash(bash_script: str):
    process = subprocess.Popen(  # noqa: S602
        bash_script,
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
        text=True,
    )

    process.wait()

    if process.returncode != 0:
        raise subprocess.CalledProcessError(process.returncode, bash_script)

    return process.returncode


def install_packages_if_missing(package_names: list[str]):
    for pkg in package_names:
        try:
            importlib.import_module(pkg)
        except ImportError:
            subprocess.check_call([sys.executable, "-m", "pip", "install", pkg])  # noqa: S603
